from opentelemetry import metrics

meter = metrics.get_meter("meshtastic_prometheus_exporter")

meshtastic_mesh_packets_total = meter.create_counter(
    name="meshtastic_mesh_packets_total",
)

meshtastic_node_info_last_heard_timestamp_seconds = meter.create_gauge(
    name="meshtastic_node_info_last_heard_timestamp_seconds",
)

meshtastic_neighbor_info_snr_decibels = meter.create_gauge(
    name="meshtastic_neighbor_info_snr_decibels",
)

meshtastic_neighbor_info_last_rx_time = meter.create_gauge(
    name="meshtastic_neighbor_info_last_rx_time",
)

meshtastic_telemetry_device_battery_level_percent = meter.create_gauge(
    name="meshtastic_telemetry_device_battery_level_percent",
)

meshtastic_telemetry_device_voltage_volts = meter.create_gauge(
    name="meshtastic_telemetry_device_voltage_volts",
)

meshtastic_telemetry_device_channel_utilization_percent = meter.create_gauge(
    name="meshtastic_telemetry_device_channel_utilization_percent",
)

meshtastic_telemetry_device_air_util_tx_percent = meter.create_gauge(
    name="meshtastic_telemetry_device_air_util_tx_percent",
)

meshtastic_telemetry_env_temperature_celsius = meter.create_gauge(
    name="meshtastic_telemetry_env_temperature_celsius",
)

meshtastic_telemetry_env_relative_humidity_percent = meter.create_gauge(
    name="meshtastic_telemetry_env_relative_humidity_percent",
)

meshtastic_telemetry_env_barometric_pressure_pascal = meter.create_gauge(
    name="meshtastic_telemetry_env_barometric_pressure_pascal",
)

meshtastic_telemetry_env_gas_resistance_ohms = meter.create_gauge(
    name="meshtastic_telemetry_env_gas_resistance_ohms",
)

meshtastic_telemetry_env_voltage_volts = meter.create_gauge(
    name="meshtastic_telemetry_env_voltage_volts",
)

meshtastic_telemetry_env_current_amperes = meter.create_gauge(
    name="meshtastic_telemetry_env_current_amperes",
)

meshtastic_telemetry_power_ch1_voltage_volts = meter.create_gauge(
    name="meshtastic_telemetry_power_ch1_voltage_volts",
)

meshtastic_telemetry_power_ch1_current_amperes = meter.create_gauge(
    name="meshtastic_telemetry_power_ch1_current_amperes",
)

meshtastic_telemetry_power_ch2_voltage_volts = meter.create_gauge(
    name="meshtastic_telemetry_power_ch2_voltage_volts",
)

meshtastic_telemetry_power_ch2_current_amperes = meter.create_gauge(
    name="meshtastic_telemetry_power_ch2_current_amperes",
)

meshtastic_telemetry_power_ch3_voltage_volts = meter.create_gauge(
    name="meshtastic_telemetry_power_ch3_voltage_volts",
)

meshtastic_telemetry_power_ch3_current_amperes = meter.create_gauge(
    name="meshtastic_telemetry_power_ch3_current_amperes",
)

meshtastic_telemetry_air_quality_pm10_standard = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_pm10_standard",
)

meshtastic_telemetry_air_quality_pm25_standard = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_pm25_standard",
)

meshtastic_telemetry_air_quality_pm100_standard = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_pm100_standard",
)

meshtastic_telemetry_air_quality_pm10_environmental = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_pm10_environmental",
)

meshtastic_telemetry_air_quality_pm25_environmental = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_pm25_environmental",
)

meshtastic_telemetry_air_quality_pm100_environmental = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_pm100_environmental",
)

meshtastic_telemetry_air_quality_particles_03um = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_particles_03um",
)

meshtastic_telemetry_air_quality_particles_05um = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_particles_05um",
)

meshtastic_telemetry_air_quality_particles_10um = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_particles_10um",
)

meshtastic_telemetry_air_quality_particles_25um = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_particles_25um",
)

meshtastic_telemetry_air_quality_particles_50um = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_particles_50um",
)

meshtastic_telemetry_air_quality_particles_100um = meter.create_gauge(
    name="meshtastic_telemetry_air_quality_particles_100um",
)



import meshtastic_prometheus_exporter.__main__ as exporter
import json
from pytest_mock import MockerFixture
from unittest.mock import call


def mocked_get_decoded_node_metadata_from_redis(redis, node: float, metadata: str):
    return "mocked"


def test_nodeinfo(mocker: MockerFixture):
    packet = '{"from": 123456789, "to": 987654321, "decoded": {"portnum": "NEIGHBORINFO_APP", "neighborinfo": {"nodeId": 123456789, "lastSentById": 123456789, "nodeBroadcastIntervalSecs": 3600, "neighbors": [{"nodeId": 123456711, "snr": 3.5}, {"nodeId": 123456722, "snr": 6.5}, {"nodeId": 123456733, "snr": -11.5}, {"nodeId": 123456744, "snr": 6.25}, {"nodeId": 123456755, "snr": 6.75}]}}, "id": 3117092156, "rxTime": 1702513724, "rxSnr": -10.0, "hopLimit": 2, "rxRssi": -119, "fromId": "!b6ffffac", "toId": "^all"}'

    packet_decoded = json.loads(packet)

    mocker.patch("meshtastic_prometheus_exporter.__main__.redis")
    mocker.patch(
        "meshtastic_prometheus_exporter.neighborinfo.get_decoded_node_metadata_from_redis",
        new=mocked_get_decoded_node_metadata_from_redis,
    )
    mocker.patch(
        "meshtastic_prometheus_exporter.__main__.get_decoded_node_metadata_from_redis",
        new=mocked_get_decoded_node_metadata_from_redis,
    )
    mock_set_last_rx_time = mocker.patch.object(
        exporter.meshtastic_neighbor_info_last_rx_time, "set"
    )
    mock_set_snr_decibels = mocker.patch.object(
        exporter.meshtastic_neighbor_info_snr_decibels, "set"
    )
    mock_add_packets_total = mocker.patch.object(
        exporter.meshtastic_mesh_packets_total, "add"
    )

    exporter.on_meshtastic_mesh_packet(packet_decoded)

    # mock_set_last_rx_time.assert_called_once()
    mock_add_packets_total.assert_called_once()

    neighbor_info = json.loads(packet)["decoded"]["neighborinfo"]
    for n in neighbor_info["neighbors"]:
        mock_set_snr_decibels.assert_called()  # TODO: more logic here



import logging
import json
from meshtastic_prometheus_exporter.metrics import *
from meshtastic_prometheus_exporter.util import get_decoded_node_metadata_from_redis

logger = logging.getLogger("meshtastic_prometheus_exporter")


def on_device_metrics_telemetry(packet, attributes):
    logger.info(f"MeshPacket {packet['id']} is device metrics telemetry")
    if "batteryLevel" in packet["decoded"]["telemetry"]["deviceMetrics"]:
        meshtastic_telemetry_device_battery_level_percent.set(
            packet["decoded"]["telemetry"]["deviceMetrics"]["batteryLevel"],
            attributes=attributes,
        )
    if "voltage" in packet["decoded"]["telemetry"]["deviceMetrics"]:
        meshtastic_telemetry_device_voltage_volts.set(
            packet["decoded"]["telemetry"]["deviceMetrics"]["voltage"],
            attributes=attributes,
        )
    if "channelUtilization" in packet["decoded"]["telemetry"]["deviceMetrics"]:
        meshtastic_telemetry_device_channel_utilization_percent.set(
            packet["decoded"]["telemetry"]["deviceMetrics"]["channelUtilization"],
            attributes=attributes,
        )
    if "airUtilTx" in packet["decoded"]["telemetry"]["deviceMetrics"]:
        meshtastic_telemetry_device_air_util_tx_percent.set(
            packet["decoded"]["telemetry"]["deviceMetrics"]["airUtilTx"],
            attributes=attributes,
        )


def on_meshtastic_telemetry_app(packet, source_long_name, source_short_name):
    telemetry = packet["decoded"]["telemetry"]
    logger.debug(
        f"Received MeshPacket {packet['id']} with Telemetry `{json.dumps(telemetry, default=repr)}`"
    )
    source = packet["decoded"].get("source", packet["from"])
    telemetry_attributes = {
        "source": source or "unknown",
        "source_long_name": source_long_name or "unknown",
        "source_short_name": source_short_name or "unknown",
    }
    if "deviceMetrics" in telemetry:
        on_device_metrics_telemetry(packet, telemetry_attributes)
        return

    if "environmentMetrics" in telemetry:
        logger.info(f"MeshPacket {packet['id']} is environment metrics telemetry")
        if "temperature" in telemetry["environmentMetrics"]:
            meshtastic_telemetry_env_temperature_celsius.set(
                telemetry["environmentMetrics"]["temperature"],
                attributes=telemetry_attributes,
            )
        if "relativeHumidity" in telemetry["environmentMetrics"]:
            meshtastic_telemetry_env_relative_humidity_percent.set(
                telemetry["environmentMetrics"]["relativeHumidity"],
                attributes=telemetry_attributes,
            )
        if "barometricPressure" in telemetry["environmentMetrics"]:
            meshtastic_telemetry_env_barometric_pressure_pascal.set(
                telemetry["environmentMetrics"]["barometricPressure"] * 10**2,
                attributes=telemetry_attributes,
            )
        if "gasResistance" in telemetry["environmentMetrics"]:
            meshtastic_telemetry_env_gas_resistance_ohms.set(
                telemetry["environmentMetrics"]["gasResistance"] / 10**6,
                attributes=telemetry_attributes,
            )
        if "voltage" in telemetry["environmentMetrics"]:
            meshtastic_telemetry_env_voltage_volts.set(
                telemetry["environmentMetrics"]["voltage"],
                attributes=telemetry_attributes,
            )
        if "current" in telemetry["environmentMetrics"]:
            meshtastic_telemetry_env_current_amperes.set(
                telemetry["environmentMetrics"]["current"] * 10**-3,
                attributes=telemetry_attributes,
            )
    if "airQualityMetrics" in telemetry:
        logger.info(f"MeshPacket {packet['id']} is air quality metrics telemetry")
        meshtastic_telemetry_air_quality_pm10_standard.set(
            telemetry["airQualityMetrics"]["pm10_standard"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_pm25_standard.set(
            telemetry["airQualityMetrics"]["pm25_standard"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_pm100_standard.set(
            telemetry["airQualityMetrics"]["pm100_standard"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_pm10_environmental.set(
            telemetry["airQualityMetrics"]["pm10_environmental"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_pm25_environmental.set(
            telemetry["airQualityMetrics"]["pm25_environmental"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_pm100_environmental.set(
            telemetry["airQualityMetrics"]["pm100_environmental"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_particles_03um.set(
            telemetry["airQualityMetrics"]["particles_03um"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_particles_05um.set(
            telemetry["airQualityMetrics"]["particles_05um"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_particles_10um.set(
            telemetry["airQualityMetrics"]["particles_10um"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_particles_25um.set(
            telemetry["airQualityMetrics"]["particles_25um"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_particles_50um.set(
            telemetry["airQualityMetrics"]["particles_50um"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_air_quality_particles_100um.set(
            telemetry["airQualityMetrics"]["particles_100um"],
            attributes=telemetry_attributes,
        )
    if "powerMetrics" in telemetry:
        logger.info(f"MeshPacket {packet['id']} is power metrics telemetry")
        meshtastic_telemetry_power_ch1_voltage_volts.set(
            telemetry["powerMetrics"]["ch1_voltage"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_power_ch1_current_amperes.set(
            telemetry["powerMetrics"]["ch1_current"] * 10**-3,
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_power_ch2_voltage_volts.set(
            telemetry["powerMetrics"]["ch2_voltage"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_power_ch2_current_amperes.set(
            telemetry["powerMetrics"]["ch2_current"] * 10**-3,
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_power_ch3_voltage_volts.set(
            telemetry["powerMetrics"]["ch3_voltage"],
            attributes=telemetry_attributes,
        )
        meshtastic_telemetry_power_ch3_current_amperes.set(
            telemetry["powerMetrics"]["ch3_current"] * 10**-3,
            attributes=telemetry_attributes,
        )



import meshtastic_prometheus_exporter.__main__ as exporter
import json
from pytest_mock import MockerFixture


def mocked_get_decoded_node_metadata_from_redis(redis, node: float, metadata: str):
    return "mocked"


def test_device_metrics_telemetry(mocker: MockerFixture):
    packet = '{"from": 123456789, "to": 987654321, "decoded": {"portnum": "TELEMETRY_APP", "telemetry": {"time": 1732550036, "deviceMetrics": {"batteryLevel": 101, "voltage": 4.122, "channelUtilization": 0.0, "airUtilTx": 0.15486111, "uptimeSeconds": 2027}}}, "id": 3259852062, "rxTime": 1732550036, "hopLimit": 3, "priority": "BACKGROUND",  "fromId": null, "toId": "^all"}'

    packet_decoded = json.loads(packet)

    mocker.patch(
        "meshtastic_prometheus_exporter.__main__.get_decoded_node_metadata_from_redis",
        new=mocked_get_decoded_node_metadata_from_redis,
    )

    mock_set_battery_level = mocker.patch.object(
        exporter.meshtastic_telemetry_device_battery_level_percent, "set"
    )
    mock_set_voltage = mocker.patch.object(
        exporter.meshtastic_telemetry_device_voltage_volts, "set"
    )
    mock_set_channel_utilization = mocker.patch.object(
        exporter.meshtastic_telemetry_device_channel_utilization_percent, "set"
    )
    mock_set_air_util = mocker.patch.object(
        exporter.meshtastic_telemetry_device_air_util_tx_percent, "set"
    )
    mock_add_packets_total = mocker.patch.object(
        exporter.meshtastic_mesh_packets_total, "add"
    )

    mocker.patch("meshtastic_prometheus_exporter.__main__.redis")
    exporter.on_meshtastic_mesh_packet(packet_decoded)

    mock_set_battery_level.assert_called_once_with(
        101,
        attributes={
            "source": packet_decoded["from"],
            "source_long_name": "mocked",
            "source_short_name": "mocked",
        },
    )
    mock_set_voltage.assert_called_once_with(
        4.122,
        attributes={
            "source": packet_decoded["from"],
            "source_long_name": "mocked",
            "source_short_name": "mocked",
        },
    )
    mock_set_channel_utilization.assert_called_once_with(
        0.0,
        attributes={
            "source": packet_decoded["from"],
            "source_long_name": "mocked",
            "source_short_name": "mocked",
        },
    )
    mock_set_air_util.assert_called_once_with(
        0.15486111,
        attributes={
            "source": packet_decoded["from"],
            "source_long_name": "mocked",
            "source_short_name": "mocked",
        },
    )
    mock_add_packets_total.assert_called_once()



import logging
import time
from meshtastic_prometheus_exporter.metrics import *
import json

from meshtastic_prometheus_exporter.util import save_node_metadata_in_redis

logger = logging.getLogger("meshtastic_prometheus_exporter")


def on_meshtastic_nodeinfo_app(redis, packet):
    node_info = packet["decoded"]["user"]

    logger.debug(
        f"Received MeshPacket {packet['id']} with NodeInfo `{json.dumps(node_info, default=repr)}`"
    )

    source = packet["decoded"].get("source", packet["from"])

    if source:
        save_node_metadata_in_redis(redis, source, node_info)

    node_info_attributes = {
        "source": source,
        "user": node_info["id"],
        "source_long_name": node_info["longName"],
        "source_short_name": node_info["shortName"],
        "is_licensed": str(node_info.get("isLicensed", 0)),
    }
    meshtastic_node_info_last_heard_timestamp_seconds.set(
        time.time(), attributes=node_info_attributes
    )



import logging
import json
from meshtastic_prometheus_exporter.util import get_decoded_node_metadata_from_redis
from meshtastic_prometheus_exporter.metrics import *

logger = logging.getLogger("meshtastic_prometheus_exporter")


def on_meshtastic_neighborinfo_app(redis, packet, source_long_name, source_short_name):
    neighbor_info = packet["decoded"]["neighborinfo"]
    logger.debug(
        f"Received MeshPacket {packet['id']} with NeighborInfo `{json.dumps(neighbor_info, default=repr)}`"
    )

    source = neighbor_info["nodeId"]
    neighbor_info_attributes = {
        "source": source,
        "source_long_name": source_long_name,
        "source_short_name": source_short_name,
    }
    for n in neighbor_info["neighbors"]:
        neighbor_source = n["nodeId"]

        neighbor_info_attributes["neighbor_source"] = neighbor_source or "unknown"
        neighbor_info_attributes["neighbor_source_long_name"] = (
            get_decoded_node_metadata_from_redis(redis, neighbor_source, "long_name")
            if source
            else "unknown"
        )
        neighbor_info_attributes["neighbor_source_short_name"] = (
            get_decoded_node_metadata_from_redis(redis, neighbor_source, "short_name")
            if source
            else "unknown"
        )

        meshtastic_neighbor_info_snr_decibels.set(
            n["snr"], attributes=neighbor_info_attributes
        )
        # https://buf.build/meshtastic/protobufs/file/main:meshtastic/mesh.proto#L1795
        # meshtastic_neighbor_info_last_rx_time.set(
        #     n["rxTime"], attributes=neighbor_info_attributes
        # )



import meshtastic_prometheus_exporter.__main__ as exporter
import json
from pytest_mock import MockerFixture


def mocked_get_decoded_node_metadata_from_redis(redis, node: float, metadata: str):
    return "mocked"


def test_nodeinfo(mocker: MockerFixture):
    packet = '{"from": 123456789, "to": 987654321, "decoded": {"portnum": "NODEINFO_APP", "bitfield": 0, "user": {"id": "!1fc44444", "longName": "namename", "shortName": "name", "macaddr": "CDIuwbaC", "hwModel": "TBEAM", "isLicensed": true}}, "id": 662811674, "rxTime": 1730000000, "rxSnr": 13.5, "hopLimit": 3, "rxRssi": -31, "hopStart": 3, "fromId": "!1fc44444", "toId": "^all"}'

    packet_decoded = json.loads(packet)

    mocker.patch("meshtastic_prometheus_exporter.__main__.redis")
    mocker.patch(
        "meshtastic_prometheus_exporter.__main__.get_decoded_node_metadata_from_redis",
        new=mocked_get_decoded_node_metadata_from_redis,
    )
    mock_set_last_heard_timestamp_seconds = mocker.patch.object(
        exporter.meshtastic_node_info_last_heard_timestamp_seconds, "set"
    )
    mock_add_packets_total = mocker.patch.object(
        exporter.meshtastic_mesh_packets_total, "add"
    )

    exporter.on_meshtastic_mesh_packet(packet_decoded)

    mock_set_last_heard_timestamp_seconds.assert_called_once()
    mock_add_packets_total.assert_called_once()



#!/usr/bin/python3

"""
meshtastic-prometheus-exporter
Copyright (C) 2024  Artiom Mocrenco and Contributors

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import logging
import os
import ssl
import sys
import time
import traceback
from sys import stdout
from google.protobuf.json_format import MessageToDict

import google.protobuf.message
import meshtastic.ble_interface
import meshtastic.serial_interface
import meshtastic.tcp_interface
import redis
from google.protobuf.json_format import MessageToDict
from opentelemetry.exporter.prometheus import PrometheusMetricReader
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.resources import Resource
import paho.mqtt.client as mqtt
from meshtastic.protobuf import mqtt_pb2
from prometheus_client import start_http_server
from pubsub import pub
from redis import BusyLoadingError
from redis.backoff import ExponentialBackoff
from redis.retry import Retry

from meshtastic_prometheus_exporter.metrics import *
from meshtastic_prometheus_exporter.neighborinfo import on_meshtastic_neighborinfo_app
from meshtastic_prometheus_exporter.nodeinfo import on_meshtastic_nodeinfo_app
from meshtastic_prometheus_exporter.telemetry import on_meshtastic_telemetry_app
from meshtastic_prometheus_exporter.util import (
    get_decoded_node_metadata_from_redis,
    save_node_metadata_in_redis,
)

config = {
    "meshtastic_interface": os.environ.get("MESHTASTIC_INTERFACE"),
    "interface_serial_device": os.environ.get("SERIAL_DEVICE", "/dev/ttyACM0"),
    "interface_tcp_addr": os.environ.get("INTERFACE_TCP_ADDR"),
    "interface_tcp_port": os.environ.get(
        "INTERFACE_TCP_PORT", meshtastic.tcp_interface.DEFAULT_TCP_PORT
    ),
    "interface_ble_addr": os.environ.get("INTERFACE_BLE_ADDR", "/dev/ttyACM0"),
    "mqtt_address": os.environ.get("MQTT_ADDRESS", "mqtt.meshtastic.org"),
    "mqtt_use_tls": os.environ.get("MQTT_USE_TLS", False),
    "mqtt_port": os.environ.get("MQTT_PORT", 1883),
    "mqtt_keepalive": os.environ.get("MQTT_KEEPALIVE", 15),
    "mqtt_username": os.environ.get("MQTT_USERNAME"),
    "mqtt_password": os.environ.get("MQTT_PASSWORD"),
    "mqtt_topic": os.environ.get("MQTT_TOPIC", "msh/EU_433/#"),
    "prometheus_server_addr": os.environ.get("PROMETHEUS_SERVER_ADDR", "0.0.0.0"),
    "prometheus_server_port": os.environ.get("PROMETHEUS_SERVER_PORT", 9464),
    "redis_url": os.environ.get("REDIS_URL", "redis://localhost:6379"),
    "log_level": os.environ.get("LOG_LEVEL", "INFO"),
    "flood_expire_time": os.environ.get("FLOOD_EXPIRE_TIME", 10 * 60),
    "enable_sentry": os.environ.get("ENABLE_SENTRY", False),
    "sentry_dsn": os.environ.get(
        "SENTRY_DSN",
        "https://d03452fcb06e7141c5c9a1d6ee370e8d@o4508362511286272.ingest.de.sentry.io/4508362517381200",
    ),
}

logger = logging.getLogger("meshtastic_prometheus_exporter")
logger.propagate = False

logger.setLevel(getattr(logging, config["log_level"].upper()))

handler = logging.StreamHandler(stdout)
handler.setFormatter(
    logging.Formatter(
        "%(asctime)s - meshtastic_prometheus_exporter - %(levelname)s - %(message)s"
    )
)

logger.addHandler(handler)

if int(config["enable_sentry"]) == 1:
    import sentry_sdk
    from sentry_sdk import add_breadcrumb
    from sentry_sdk.integrations.logging import LoggingIntegration

    logger.info(
        "Enabling error reporting via Sentry. Your unmodified MeshPackets will be sent to maintainers of this project in case of runtime errors. To disable, set ENABLE_SENTRY environment variable to 0"
    )

    def before_breadcrumb(crumb, hint):
        if crumb["category"] == "redis":
            return None
        return crumb

    sentry_sdk.init(
        dsn=config.get("sentry_dsn"),
        traces_sample_rate=1.0,
        before_breadcrumb=before_breadcrumb,
        integrations=[
            LoggingIntegration(level=logging.INFO, event_level=logging.FATAL),
        ],
    )
else:
    logger.info(
        "Error reporting via Sentry is disabled. If you want to send your unmodified MeshPackets to maintainers of this project in case of runtime errors, set ENABLE_SENTRY environment variable to 1"
    )


try:
    reader = PrometheusMetricReader()
    start_http_server(
        port=config["prometheus_server_port"], addr=config["prometheus_server_addr"]
    )

    provider = MeterProvider(
        resource=Resource.create(attributes={"service.name": "meshtastic"}),
        metric_readers=[reader],
        # views=[],
    )
    metrics.set_meter_provider(provider)
    meter = metrics.get_meter("meshtastic_prometheus_exporter")

    redis = redis.from_url(
        config["redis_url"],
        db=0,
        protocol=3,
        retry=Retry(ExponentialBackoff(10, 1), 3),
        retry_on_error=[BusyLoadingError, ConnectionError, TimeoutError],
    )

except Exception as e:
    logger.fatal(
        f"Exception occurred while starting up: {';'.join(traceback.format_exc().splitlines())}"
    )
    sys.exit(1)


def on_connect(client, userdata, flags, reason_code, properties):
    if reason_code.is_failure:
        logger.warning(
            f"Failed to connect to MQTT server with result code {reason_code}. loop_forever() will retry connection"
        )
    else:
        logger.info(f"Connected to MQTT server with result code {reason_code}")
        client.subscribe(config["mqtt_topic"])


def on_meshtastic_service_envelope(envelope, msg):
    logger.debug(
        f"Received ServiceEnvelope payload `{msg['payload']}` from `{msg.topic}` topic"
    )

    if envelope:
        logger.debug(
            f"Decoded ServiceEnvelope payload into MeshPacket `{envelope.packet}`"
        )
        on_meshtastic_mesh_packet(envelope.packet)


def on_meshtastic_mesh_packet(packet):
    if packet.get("encrypted", False):
        logger.info(f"Skipping encrypted packet {packet['id']}")
        return

    if packet.get("id", None) is None:
        return

    unique = redis.set(str(packet["id"]), 1, nx=True, ex=config["flood_expire_time"])

    if not unique:
        logger.info(f"Skipping duplicate packet {packet['id']}")
        return

    source = packet["decoded"].get("source", packet["from"])

    source_long_name = get_decoded_node_metadata_from_redis(redis, source, "long_name")
    source_short_name = get_decoded_node_metadata_from_redis(
        redis, source, "short_name"
    )
    from_long_name = get_decoded_node_metadata_from_redis(
        redis, packet["from"], "long_name"
    )
    from_short_name = get_decoded_node_metadata_from_redis(
        redis, packet["from"], "short_name"
    )
    to_long_name = get_decoded_node_metadata_from_redis(
        redis, packet["to"], "long_name"
    )
    to_short_name = get_decoded_node_metadata_from_redis(
        redis, packet["to"], "short_name"
    )

    # https://buf.build/meshtastic/protobufs/file/main:meshtastic/portnums.proto
    meshtastic_mesh_packets_total.add(
        1,
        attributes={
            "source": source,
            "source_long_name": source_long_name,
            "source_short_name": source_short_name,
            "from": packet["from"],
            "from_long_name": from_long_name,
            "from_short_name": from_short_name,
            "to": packet["to"],
            "to_long_name": to_long_name,
            "to_short_name": to_short_name,
            "channel": packet.get("channel", 0),
            "type": packet["decoded"]["portnum"],
            "hop_limit": packet.get("hopLimit", "unknown"),
            "want_ack": packet.get("wantAck", "unknown"),
            "delayed": packet.get("delayed", "unknown"),
            "via_mqtt": packet.get("viaMqtt", "false"),
        },
    )
    if packet["decoded"]["portnum"] == "NODEINFO_APP":
        on_meshtastic_nodeinfo_app(redis, packet)
    else:
        known_source = (
            get_decoded_node_metadata_from_redis(redis, source, "long_name")
            != "unknown"
        )

        if not known_source:
            logger.info(
                f"NodeInfo is now yet known for Node {source}, ignoring the packet {packet['id']}"
            )
            return

    if packet["decoded"]["portnum"] == "TELEMETRY_APP":
        on_meshtastic_telemetry_app(packet, source_long_name, source_short_name)

    if packet["decoded"]["portnum"] == "NEIGHBORINFO_APP":
        on_meshtastic_neighborinfo_app(
            redis, packet, source_long_name, source_short_name
        )


def on_message(client, userdata, msg):
    try:
        envelope = mqtt_pb2.ServiceEnvelope.FromString(msg.payload)
        packet = envelope.packet
        logger.debug(
            f"Received UTF-8 payload `{MessageToDict(envelope)}` from `{msg.topic}` topic"
        )
        on_native_message(MessageToDict(packet), None)
    except Exception as e:
        logger.warning(f"Exception occurred in on_message: {e}")


def on_native_message(packet, interface):
    try:
        on_meshtastic_mesh_packet(packet)
    except Exception as e:
        logger.error(
            f"{e} occurred while processing MeshPacket {packet}, please consider submitting a PR/issue on GitHub: `{json.dumps(packet, default=repr)}` {';'.join(traceback.format_exc().splitlines())
}"
        )
        if "sentry_sdk" in globals():
            sentry_sdk.capture_exception(e)


def on_native_connection_established(interface, topic=pub.AUTO_TOPIC):
    logger.info(f"Connected to device over {type(interface).__name__}")


def on_native_connection_lost(interface, topic=pub.AUTO_TOPIC):
    logger.warning(f"Lost connection to device over {type(interface).__name__}")


def main():
    try:
        logger.info(
            "Share ideas and vote for new features https://github.com/artiommocrenco/meshtastic-prometheus-exporter/discussions/categories/ideas"
        )

        if config.get("meshtastic_interface") not in ["MQTT", "SERIAL", "TCP", "BLE"]:
            logger.fatal(
                f"Invalid value for MESHTASTIC_INTERFACE: {config['meshtastic_interface']}. Must be one of: MQTT, SERIAL, TCP, BLE"
            )
            sys.exit(1)

        pub.subscribe(on_native_message, "meshtastic.receive")
        pub.subscribe(
            on_native_connection_established, "meshtastic.connection.established"
        )
        pub.subscribe(on_native_connection_lost, "meshtastic.connection.lost")

        if config.get("meshtastic_interface") == "SERIAL":
            iface = meshtastic.serial_interface.SerialInterface(
                devPath=config.get("serial_device")
            )
        elif config.get("meshtastic_interface") == "TCP":
            iface = meshtastic.tcp_interface.TCPInterface(
                hostname=config.get("interface_tcp_addr"),
                portNumber=int(config.get("interface_tcp_port")),
            )
        elif config.get("meshtastic_interface") == "BLE":
            iface = meshtastic.ble_interface.BLEInterface(
                address=config.get("interface_ble_addr"),
            )
        elif config.get("meshtastic_interface") == "MQTT":
            mqttc = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)

            mqttc.on_connect = on_connect
            mqttc.on_message = on_message

            if int(config["mqtt_use_tls"]) == 1:
                tlscontext = ssl.create_default_context()
                mqttc.tls_set_context(tlscontext)

            if config["mqtt_username"]:
                mqttc.username_pw_set(config["mqtt_username"], config["mqtt_password"])

            mqttc.connect(
                config["mqtt_address"],
                int(config["mqtt_port"]),
                keepalive=int(config["mqtt_keepalive"]),
            )

            mqttc.loop_forever()

        if hasattr(iface, "nodes") and len(iface.nodes) > 0:
            logger.info(
                f"NodeDB is available, saving metadata in Redis for {len(iface.nodes.values())} nodes"
            )
            for n in iface.nodes.values():
                save_node_metadata_in_redis(
                    redis,
                    n["num"],
                    {
                        "longName": n["user"]["longName"],
                        "shortName": n["user"]["shortName"],
                        "hwModel": n["user"]["hwModel"],
                    },
                )
        else:
            logger.warning(
                "Device NodeDB is empty or not available. NodeInfo are not sent often, so populating local NodeDB (stored in Redis) may take from several hours to several days or more."
            )
            logger.warning(
                "Consider first connecting a node with populated NodeDB over Serial, BLE or TCP interface, so that Redis is populated with NodeInfo faster."
            )

        while True:
            time.sleep(1)

    except Exception as e:
        logger.fatal(
            f"Exception occurred while starting up: {';'.join(traceback.format_exc().splitlines())}"
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



def get_decoded_node_metadata_from_redis(redis, node: float, metadata: str):
    v = redis.hget(node, metadata)
    try:
        v = "unknown" if v is None else v.decode("utf-8")
    except UnicodeDecodeError:
        v = "unknown"
    return v


def save_node_metadata_in_redis(redis, node: float, node_info: dict, ex=3600 * 72):
    redis.hset(
        node,
        mapping={
            "long_name": node_info["longName"],
            "short_name": node_info["shortName"],
            "hw_model": node_info["hwModel"],
            "is_licensed": str(node_info.get("isLicensed", False)),
        },
    )
    redis.expire(node, ex)
